pytest_plugins = [
    "tests.fixtures.logger_fixture",
    "tests.fixtures.repository_fixture",
    "tests.fixtures.restic_cli_fixture",
    "tests.fixtures.snapshot_fixture",
    "tests.fixtures.snapshot_list_fixture",
]
